@@ -7,10 +7,11 @@
 from enum import IntEnum
 from functools import wraps
 from hashlib import sha256
-from io import BufferedIOBase
+from io import RawIOBase
 from os import environ
 from os.path import getsize, join as pjoin
 from struct import pack, unpack
+from threading import Lock
 from typing import TYPE_CHECKING
 
 from Cryptodome.Cipher import AES
@@ -31,8 +32,18 @@ if TYPE_CHECKING:
     from Cryptodome.Hash.CMAC import CMAC as CMAC_CLASS
     from typing import BinaryIO, Dict, List, Optional, Union
 
-__all__ = ['CryptoError', 'OTPLengthError', 'CorruptBootromError', 'KeyslotMissingError', 'TicketLengthError',
-           'BootromNotFoundError', 'CorruptOTPError', 'Keyslot', 'CryptoEngine', 'CTRFileIO', 'CBCFileIO']
+    # trick type checkers
+    RawIOBase = BinaryIO
+
+__all__ = ['MIN_TICKET_SIZE', 'CryptoError', 'OTPLengthError', 'CorruptBootromError', 'KeyslotMissingError',
+           'TicketLengthError', 'BootromNotFoundError', 'CorruptOTPError', 'Keyslot', 'CryptoEngine', 'CTRFileIO',
+           'CBCFileIO']
+
+BOOT9_PROT_HASH = '7331f7edece3dd33f2ab4bd0b3a5d607229fd19212c10b734cedcaf78c1a7b98'
+
+DEV_COMMON_KEY_0 = bytes.fromhex('55A3F872BDC80C555A654381139E153B')
+
+MIN_TICKET_SIZE = 0x2AC
 
 
 class CryptoError(PyCTRError):
@@ -74,30 +85,49 @@ class CorruptBootromError(CryptoError):
 
 
 class Keyslot(IntEnum):
+    """
+    AES engine keyslots used by the Nintendo 3DS. Values above 0x3F (63) are used by PyCTR, and do not exist on the
+    actual hardware. Each value explains what the keyslot is used to decrypt or encrypt.
+    """
+
     TWLNAND = 0x03
+    """Entire TWL region, including twln, twlp, and the header."""
+
     CTRNANDOld = 0x04
+    """CTRNAND for Old Nintendo 3DS."""
     CTRNANDNew = 0x05
+    """CTRNAND for New Nintendo 3DS."""
     FIRM = 0x06
+    """FIRM partitions."""
     AGB = 0x07
+    """AGBSAVE partition if a GBA VC title was played."""
 
     CMACNANDDB = 0x0B
+    """CMAC for NAND dbs."""
 
     NCCH93 = 0x18
+    """NCCH extra keyslot for titles exclusive to New Nintendo 3DS released after System Menu 9.3.0-21."""
     CMACCardSaveNew = 0x19
     CardSaveNew = 0x1A
     NCCH96 = 0x1B
+    """NCCH extra keyslot for titles exclusive to New Nintendo 3DS released after System Menu 9.6.0-24."""
 
     CMACAGB = 0x24
+    """CMAC for the AGBSAVE partition contents."""
     NCCH70 = 0x25
+    """NCCH extra keyslot for titles released after System Menu 7.0.0-13."""
 
     NCCH = 0x2C
-    UDSLocalWAN = 0x2D
+    """NCCH original keyslot."""
+    UDSLocalWLAN = 0x2D
     StreetPass = 0x2E
     Save60 = 0x2F
+    """Save key for retail games released after System Menu 6.0.0-11."""
     CMACSDNAND = 0x30
 
     CMACCardSave = 0x33
     SD = 0x34
+    """SD card contents under "Nintendo 3DS"."""
 
     CardSave = 0x37
     BOSS = 0x38
@@ -106,15 +136,13 @@ class Keyslot(IntEnum):
     DSiWareExport = 0x3A
 
     CommonKey = 0x3D
+    """Titlekeys in tickets."""
 
     # anything after 0x3F is custom to PyCTR
     DecryptedTitlekey = 0x40
+    """CIA and CDN contents."""
 
 
-BOOT9_PROT_HASH = '7331f7edece3dd33f2ab4bd0b3a5d607229fd19212c10b734cedcaf78c1a7b98'
-
-DEV_COMMON_KEY_0 = bytes.fromhex('55A3F872BDC80C555A654381139E153B')
-
 common_key_y = (
     # eShop
     0xD07B337F9CA4385932A2E25723232EB9,
@@ -204,17 +232,20 @@ class CryptoEngine:
     :param setup_b9_keys: Whether or not to automatically load keys from boot9.
     """
 
-    b9_keys_set: bool = False
-    b9_path: str = None
+    __slots__ = ['key_x', 'key_y', 'key_normal', 'dev', 'b9_keys_set', 'b9_path', '_b9_extdata_otp',
+                 '_b9_extdata_keygen', '_otp_device_id', '_otp_key', '_otp_iv', '_id0']
 
-    _b9_extdata_otp: bytes = None
-    _b9_extdata_keygen: bytes = None
+    b9_keys_set: bool
+    """Keys have been set from the ARM9 BootROM."""
 
-    _otp_device_id: int = None
-    _otp_key: bytes = None
-    _otp_iv: bytes = None
+    b9_path: 'Optional[str]'
+    """
+    Path that the ARM9 BootROM was loaded from. Set when :meth:`setup_keys_from_boot9_file` is called, which is
+    automatically called on object creation if `setup_b9_keys` was `True`.
+    """
 
-    _id0: bytes = None
+    dev: bool
+    """Uses devunit keys."""
 
     def __init__(self, boot9: str = None, dev: bool = False, setup_b9_keys: bool = True):
         self.key_x: Dict[int, int] = {}
@@ -224,6 +255,18 @@ class CryptoEngine:
 
         self.dev = dev
 
+        self.b9_keys_set = False
+        self.b9_path = None
+
+        self._b9_extdata_otp: Optional[bytes] = None
+        self._b9_extdata_keygen: Optional[bytes] = None
+
+        self._otp_device_id: Optional[int] = None
+        self._otp_key: Optional[bytes] = None
+        self._otp_iv: Optional[bytes] = None
+
+        self._id0: Optional[bytes] = None
+
         for keyslot, keys in base_key_x.items():
             self.key_x[keyslot] = keys[dev]
 
@@ -373,6 +416,18 @@ class CryptoEngine:
         hash_p2 = readbe(path_hash[16:32])
         return hash_p1 ^ hash_p2
 
+    def load_encrypted_titlekey(self, titlekey: bytes, common_key_index: int, title_id: 'Union[str, bytes]'):
+        if isinstance(title_id, str):
+            title_id = bytes.fromhex(title_id)
+
+        if self.dev and common_key_index == 0:
+            self.set_normal_key(0x3D, DEV_COMMON_KEY_0)
+        else:
+            self.set_keyslot('y', 0x3D, common_key_y[common_key_index])
+
+        cipher = self.create_cbc_cipher(Keyslot.CommonKey, title_id + (b'\0' * 8))
+        self.set_normal_key(0x40, cipher.decrypt(titlekey))
+
     def load_from_ticket(self, ticket: bytes):
         """Load a titlekey from a ticket and set keyslot 0x40 to the decrypted titlekey."""
         ticket_len = len(ticket)
@@ -385,13 +440,7 @@ class CryptoEngine:
         title_id = ticket[0x1DC:0x1E4]
         common_key_index = ticket[0x1F1]
 
-        if self.dev and common_key_index == 0:
-            self.set_normal_key(0x3D, DEV_COMMON_KEY_0)
-        else:
-            self.set_keyslot('y', 0x3D, common_key_y[common_key_index])
-
-        cipher = self.create_cbc_cipher(Keyslot.CommonKey, title_id + (b'\0' * 8))
-        self.set_normal_key(0x40, cipher.decrypt(titlekey_enc))
+        self.load_encrypted_titlekey(titlekey_enc, common_key_index, title_id)
 
     def set_keyslot(self, xy: str, keyslot: int, key: 'Union[int, bytes]'):
         """Sets a keyslot to the specified key."""
@@ -665,7 +714,7 @@ class CryptoEngine:
             self.setup_sd_key(f.read(0x140))
 
 
-class _CryptoFileBase(BufferedIOBase):
+class _CryptoFileBase(RawIOBase):
     """Base class for CTR and CBC IO classes."""
 
     closed = False
@@ -709,30 +758,31 @@ class CTRFileIO(_CryptoFileBase):
         self._crypto = crypto
         self._keyslot = keyslot
         self._counter = counter
+        self._lock = Lock()
 
     def __repr__(self):
         return f'{type(self).__name__}(file={self._reader!r}, keyslot={self._keyslot:#04x}, counter={self._counter!r})'
 
     @_raise_if_file_closed
     def read(self, size: int = -1) -> bytes:
-        cur_offset = self.tell()
-        data = self._reader.read(size)
-        counter = self._counter + (cur_offset >> 4)
-        cipher = self._crypto.create_ctr_cipher(self._keyslot, counter)
-        # beginning padding
-        cipher.decrypt(b'\0' * (cur_offset % 0x10))
-        return cipher.decrypt(data)
-
-    read1 = read  # probably make this act like read1 should, but this for now enables some other things to work
+        with self._lock:
+            cur_offset = self.tell()
+            data = self._reader.read(size)
+            counter = self._counter + (cur_offset >> 4)
+            cipher = self._crypto.create_ctr_cipher(self._keyslot, counter)
+            # beginning padding
+            cipher.decrypt(b'\0' * (cur_offset % 0x10))
+            return cipher.decrypt(data)
 
     @_raise_if_file_closed
     def write(self, data: bytes) -> int:
-        cur_offset = self.tell()
-        counter = self._counter + (cur_offset >> 4)
-        cipher = self._crypto.create_ctr_cipher(self._keyslot, counter)
-        # beginning padding
-        cipher.encrypt(b'\0' * (cur_offset % 0x10))
-        return self._reader.write(cipher.encrypt(data))
+        with self._lock:
+            cur_offset = self.tell()
+            counter = self._counter + (cur_offset >> 4)
+            cipher = self._crypto.create_ctr_cipher(self._keyslot, counter)
+            # beginning padding
+            cipher.encrypt(b'\0' * (cur_offset % 0x10))
+            return self._reader.write(cipher.encrypt(data))
 
     @_raise_if_file_closed
     def seek(self, seek: int, whence: int = 0) -> int:
@@ -751,50 +801,51 @@ class CBCFileIO(_CryptoFileBase):
         self._crypto = crypto
         self._keyslot = keyslot
         self._iv = iv
+        self._lock = Lock()
 
     def __repr__(self):
         return f'{type(self).__name__}(file={self._reader!r}, keyslot={self._keyslot:#04x}, iv={self._iv!r})'
 
     @_raise_if_file_closed
     def read(self, size: int = -1):
-        offset = self._reader.tell()
+        with self._lock:
+            offset = self._reader.tell()
 
-        # if encrypted, the block needs to be decrypted first
-        # CBC requires a full block (0x10 in this case). and the previous
-        #   block is used as the IV. so that's quite a bit to read if the
-        #   application requires just a few bytes.
-        # thanks Stary2001 for help with random-access crypto
+            # if encrypted, the block needs to be decrypted first
+            # CBC requires a full block (0x10 in this case). and the previous
+            #   block is used as the IV. so that's quite a bit to read if the
+            #   application requires just a few bytes.
+            # thanks Stary2001 for help with random-access crypto
 
-        before = offset % 16
-        if offset - before == 0:
-            iv = self._iv
-        else:
-            # seek back one block to read it as iv
-            self._reader.seek(-0x10 - before, 1)
-            iv = self._reader.read(0x10)
-        # this is done since we may not know the original size of the file
-        # and the caller may have requested -1 to read all the remaining data
-        data_before = self._reader.read(before)
-        data_requested = self._reader.read(size)
-        data_requested_len = len(data_requested)
-        data_total_len = len(data_before) + data_requested_len
-        if data_total_len % 16:
-            data_after = self._reader.read(16 - (data_total_len % 16))
-            self._reader.seek(-len(data_after), 1)
-        else:
-            data_after = b''
-        cipher = self._crypto.create_cbc_cipher(self._keyslot, iv)
-        # decrypt data, and cut off extra bytes
-        return cipher.decrypt(
-            b''.join((data_before, data_requested, data_after))
-        )[before:data_requested_len + before]
-
-    read1 = read  # probably make this act like read1 should, but this for now enables some other things to work
+            before = offset % 16
+            if offset - before == 0:
+                iv = self._iv
+            else:
+                # seek back one block to read it as iv
+                self._reader.seek(-0x10 - before, 1)
+                iv = self._reader.read(0x10)
+            # this is done since we may not know the original size of the file
+            # and the caller may have requested -1 to read all the remaining data
+            data_before = self._reader.read(before)
+            data_requested = self._reader.read(size)
+            data_requested_len = len(data_requested)
+            data_total_len = len(data_before) + data_requested_len
+            if data_total_len % 16:
+                data_after = self._reader.read(16 - (data_total_len % 16))
+                self._reader.seek(-len(data_after), 1)
+            else:
+                data_after = b''
+            cipher = self._crypto.create_cbc_cipher(self._keyslot, iv)
+            # decrypt data, and cut off extra bytes
+            return cipher.decrypt(
+                b''.join((data_before, data_requested, data_after))
+            )[before:data_requested_len + before]
 
     @_raise_if_file_closed
     def seek(self, seek: int, whence: int = 0):
         # even though read re-seeks to read required data, this allows the underlying object to handle seek how it wants
-        return self._reader.seek(seek, whence)
+        with self._lock:
+            return self._reader.seek(seek, whence)
 
     @_raise_if_file_closed
     def writable(self) -> bool:
