@@ -4,6 +4,8 @@
 # This file is licensed under The MIT License (MIT).
 # You can find the full license text in LICENSE in the root of this project.
 
+"""Module for interacting with NCCH files."""
+
 from hashlib import sha256
 from enum import IntEnum
 from math import ceil
@@ -12,7 +14,7 @@ from threading import Lock
 from typing import TYPE_CHECKING, NamedTuple
 
 from ..common import PyCTRError, _ReaderOpenFileBase
-from ..crypto import CryptoEngine, Keyslot, get_seed
+from ..crypto import CryptoEngine, Keyslot, add_seed, get_seed
 from ..fileio import SubsectionIO
 from ..util import readle, roundup
 from .base import TypeReaderCryptoBase
@@ -151,6 +153,7 @@ class NCCHReader(TypeReaderCryptoBase):
     :param crypto: A custom :class:`crypto.CryptoEngine` object to be used. Defaults to None, which causes a new one to
         be created.
     :param dev: Use devunit keys.
+    :param seed: Seed to use. This is a quick way to add a seed using :func:`~.seeddb.add_seed`.
     :param load_sections: Load the ExeFS and RomFS as :class:`type.exefs.ExeFSReader` and
         :class:`type.romfs.RomFSReader` objects.
     :param assume_decrypted: Assume each NCCH content is decrypted. Needed if the image was decrypted without fixing
@@ -194,7 +197,7 @@ class NCCHReader(TypeReaderCryptoBase):
     """NCCH version. Not to be confused with the title version."""
 
     def __init__(self, file: 'Union[PathLike, str, bytes, BinaryIO]', *, closefd: bool = None,
-                 case_insensitive: bool = True, crypto: CryptoEngine = None, dev: bool = False,
+                 case_insensitive: bool = True, crypto: CryptoEngine = None, dev: bool = False, seed: bytes = None,
                  load_sections: bool = True, assume_decrypted: bool = False):
 
         super().__init__(file, closefd=closefd, crypto=crypto, dev=dev)
@@ -282,6 +285,9 @@ class NCCHReader(TypeReaderCryptoBase):
         # tells if the right seed has been set up
         self._seed_set_up = False
 
+        if seed:
+            add_seed(self.program_id, seed)
+
         # load the seed if needed
         if self.flags.uses_seed:
             self.setup_seed(get_seed(self.program_id))
@@ -293,6 +299,15 @@ class NCCHReader(TypeReaderCryptoBase):
         if load_sections:
             self.load_sections()
 
+    def __repr__(self):
+        info = [('program_id', self.program_id), ('product_code', self.product_code)]
+        try:
+            info.append(('title_name', repr(self.exefs.icon.get_app_title().short_desc)))
+        except (KeyError, AttributeError):
+            info.append(('title_name', 'unknown'))
+        info_final = " ".join(x + ": " + str(y) for x, y in info)
+        return f'<{type(self).__name__} {info_final}>'
+
     def load_sections(self):
         """Load the sections of the NCCH (Extended Header, ExeFS, and RomFS)."""
 
